/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include "ScatterWeb.System.h"

#define CC1020_MAIN            0x00
#define CC1020_INTERFACE       0x01
#define CC1020_RESET           0x02
#define CC1020_SEQUENCING      0x03
#define CC1020_FREQ_2A         0x04
#define CC1020_FREQ_1A         0x05
#define CC1020_FREQ_0A         0x06
#define CC1020_CLOCK_A         0x07
#define CC1020_FREQ_2B         0x08
#define CC1020_FREQ_1B         0x09
#define CC1020_FREQ_0B         0x0A
#define CC1020_CLOCK_B         0x0B
#define CC1020_VCO             0x0C
#define CC1020_MODEM           0x0D
#define CC1020_DEVIATION       0x0E
#define CC1020_AFC_CONTROL     0x0F
#define CC1020_FILTER          0x10
#define CC1020_VGA1            0x11
#define CC1020_VGA2            0x12
#define CC1020_VGA3            0x13
#define CC1020_VGA4            0x14
#define CC1020_LOCK            0x15
#define CC1020_FRONTEND        0x16
#define CC1020_ANALOG          0x17
#define CC1020_BUFF_SWING      0x18
#define CC1020_BUFF_CURRENT    0x19
#define CC1020_PLL_BW          0x1A
#define CC1020_CALIBRATE       0x1B
#define CC1020_PA_POWER        0x1C
#define CC1020_MATCH           0x1D
#define CC1020_PHASE_COMP      0x1E
#define CC1020_GAIN_COMP       0x1F
#define CC1020_POWERDOWN       0x20
#define CC1020_TEST1           0x21
#define CC1020_TEST2           0x22
#define CC1020_TEST3           0x23
#define CC1020_TEST4           0x24
#define CC1020_TEST5           0x25
#define CC1020_TEST6           0x26
#define CC1020_TEST7           0x27
#define CC1020_STATUS          0x40
#define CC1020_RESET_DONE      0x41
#define CC1020_AFC             0x43
#define CC1020_GAUSS_FILTER    0x44
#define CC1020_STATUS1         0x45
#define CC1020_STATUS2         0x46
#define CC1020_STATUS3         0x47
#define CC1020_STATUS4         0x48
#define CC1020_STATUS5         0x49
#define CC1020_STATUS6         0x4A
#define CC1020_STATUS7         0x4B

// For CC1020_STATUS
#define LOCK_CONTINOUS (0x10)
#define CAL_COMPLETE   (0x80)
#define PA_POWER       0x05
#define LOCK_NOK       0x00
#define LOCK_OK        0x01
#define LOCK_RECAL_OK  0x02
#define CAL_TIMEOUT    0x7FFE
#define LOCK_TIMEOUT   0x7FFE
#define RESET_TIMEOUT  0x7FFE
#define PREAMBLE  0xAA

#define PDO            (P2IN & 0x01)                    // PDI (Data in) is on P21
#define PSEL_ON        do { P3OUT &=  ~0x01; } while(0) // PSEL is on P30 and low active
#define PSEL_OFF       do { P3OUT |=   0x01; } while(0)
#define PCLK_HIGH      do { P2OUT |=  0x08; }  while(0)
#define PCLK_LOW       do { P2OUT &= ~0x08; }  while(0)
#define PDI_HIGH       do { P2OUT |=  0x02; }  while(0) // PDO (Data out) is on P22
#define PDI_LOW        do { P2OUT &= ~0x02; }  while(0)
#define LNA_POWER_ON   do { P2OUT &= ~0x10; }  while(0) // Enable power for LNA (P24, low-active)

UINT16 tb1;
UINT16 tb2;

#define ACK_TIMEOUT_115   (4)  // In RADIO_STROKE ticks
#define ACK_TIMEOUT_19    (16)

typedef struct {
  UINT8 registerValues[32];
  UINT8 ackTimeout;
  UINT8 unused1;
  UINT8 unused2;
  UINT8 unused3;
  UINT8 name[28];
} chipconConfig_t;

chipconConfig_t currentCC1020Config;

#define MHZ_869525      (1)

const UINT8* currentConfig;

const UINT8 DefaultConfig_19200[41] = {
  0x01,   // 0x00, MAIN
  0x0F,   // 0x01, INTERFACE
  0xFF,   // 0x02, RESET
  0x8F,   // 0x03, SEQUENCING
  //869.525 bei 50kHz
  0x3A,   // 0x04, FREQ_2A
  0x32,   // 0x05, FREQ_1A
  0x97,   // 0x06, FREQ_0A // 19200
  0x38,   // 0x07, CLOCK_A // 19200
  0x3A,   // 0x08, FREQ_2B
  0x37,   // 0x09, FREQ_1B
  0xEB,   // 0x0A, FREQ_0B // 19200
  0x38,   // 0x0B, CLOCK_B // 19200
  0x44,   // 0x0C, VCO     44
  0x51,   // 0x0D, MODEM   Manchester
  0x2B,   // 0x0E, DEVIATION // FSK
  0x4C,   // 0x0F, AFC_CONTROL Ruetten 0xCC
  0x25,   // 0x10, FILTER Bandwith 100 kHz
  0x61,   // 0x11, VGA1
  0x55,   // 0x12, VGA2
  0x2D,   // 0x13, VGA3
  0x37,   //0x29,  // 0x14, VGA4 ADJUSTED CS to 23!
  0x40,   // 0x15, LOCK is Carrier SENSE
  0x76,   // 0x16, FRONTEND
  0x87,   // 0x17, ANALOG, RX=86/TX=87
  0x10,   // 0x18, BUFF_SWING
  0x25,   // 0x19, BUFF_CURRENT
  0xAE,   // 0x1A, PLL_BW
  0x34,   // 0x1B, CALIBRATE
  PA_POWER, // 0x1C, PA_POWER AN025 = 0xA0
  0xF0,   // 0x1D, MATCH
  0x00,   // 0x1E, PHASE_COMP
  0x00,   // 0x1F, GAIN_COMP
  0x00,   // 0x20, POWERDOWN
  0x00,   // 0x4d,  // 0x21,
  0x00,   // 0x10,  // 0x22,
  0x00,   // 0x06,  // 0x23,
  0x00,   // 0x00,  // 0x24,
  0x00,   // 0x40,  // 0x25,
  0x00,   // 0x00,  // 0x26,
  0x00,   // 0x00,   // 0x27,
  ACK_TIMEOUT_19  // Not in real config of chipCon from here!!!                 
};

const UINT8 DefaultConfig_115200[41] = 
{
  0x01,   // 0x00, MAIN
  0x0F,   // 0x01, INTERFACE
  0xFF,   // 0x02, RESET
  0x8F,   // 0x03, SEQUENCING
  // 869.525 bei 200kHz
  0x3A,   // 0x04, FREQ_2A
  0x32,   // 0x05, FREQ_1A
  0x97,   // 0x06, FREQ_0A // 19200
  0x29,   // 0x07, CLOCK_A // 19200
  0x3A,   // 0x08, FREQ_2B
  0x37,   // 0x09, FREQ_1B
  0xEB,   // 0x0A, FREQ_0B // 19200
  0x29,   // 0x0B, CLOCK_B // 19200
  0x44,   // 0x0C, VCO 44
  0x51,   // 0x0D, MODEM Manchester
  0x58,   // 0x0E, DEVIATION // FSK
  0x4C,   // 0x0F, AFC_CONTROL Ruetten 0xCC
  0x80,   // 0x10, FILTER Bandwith 500 kHz
  0x61,   // 0x11, VGA1
  0x55,   // 0x12, VGA2
  0x30,   // 0x13, VGA3
  0x35,   // 0x14, VGA4
  0x20,   // 0x15, LOCK is Carrier SENSE
  0x76,   // 0x16, FRONTEND
  0x87,   // 0x17, ANALOG, RX=86/TX=87
  0x10,   // 0x18, BUFF_SWING
  0x25,   // 0x19, BUFF_CURRENT
  0xAE,   // 0x1A, PLL_BW
  0x34,   // 0x1B, CALIBRATE
  PA_POWER, // 0x1C, PA_POWER AN025 = 0xA0
  0xF0,   // 0x1D, MATCH
  0x00,   // 0x1E, PHASE_COMP
  0x00,   // 0x1F, GAIN_COMP
  0x00,   // 0x20, POWERDOWN
  0x4d,   // 0x21,
  0x10,   // 0x22,
  0x06,   // 0x23,
  0x00,   // 0x24,
  0x40,   // 0x25,
  0x00,   // 0x26,
  0x00,   // 0x27,             
  ACK_TIMEOUT_115  // Not in real config of chipCon from here!!!           
};

void CC1020_writeRegister(UINT8 addr, UINT8 adata) {
  char  BitCounter;
  unsigned char Data ;
  PSEL_OFF;
  Data = addr << 1;
  PSEL_ON;
  // Send address bits 
  for (BitCounter=0;BitCounter<7;BitCounter++) {
    nop();         // Delay50ns();
  	PCLK_LOW;
    nop();         // Delay50ns();
    if(Data & 0x80)  PDI_HIGH;
    else PDI_LOW;
    Data=Data<<1;
    PCLK_HIGH;
  }
  // Send read/write bit 
  // Ignore bit in data, always use 1 
  nop();           // Delay50ns();
  PCLK_LOW;
  PDI_HIGH;
  nop();           // Delay50ns();
  PCLK_HIGH;
  nop();           // Delay50ns();
  PCLK_LOW;
  Data=adata ;    
  // Send data bits 
  for (BitCounter=0;BitCounter<8;BitCounter++) {
    nop();         // Delay50ns();
  	PCLK_LOW;
    nop();         // Delay50ns();
    if(Data & 0x80) PDI_HIGH;
    else PDI_LOW;
    Data=Data<<1;
    PCLK_HIGH;
  }
  nop();           // Delay50ns();
  PCLK_LOW;
  nop();           // Delay50ns();
  PSEL_OFF;
}

UINT8 CC1020_readRegister(UINT8 addr) {
  char BitCounter;
  unsigned char Data = 0;
  PSEL_OFF; 
  Data=addr << 1;
  PSEL_ON;
  nop();           // Delay50ns();
  // Send address bits 
  for (BitCounter=0;BitCounter<7;BitCounter++) {
    nop();           // Delay50ns();
  	PCLK_LOW;
    nop();           // Delay50ns();
    if(Data & 0x80)  PDI_HIGH;
    else PDI_LOW;
    Data=Data<<1;
    PCLK_HIGH;
  }
  // Send read/write bit 
  // Ignore bit in data, always use 0 
  nop();             // Delay50ns();
  PCLK_LOW;
  PDI_LOW;
  nop();             // Delay50ns();
  PCLK_HIGH;
  nop();             // Delay50ns();
  PCLK_LOW;
  // Receive data bits       
  for (BitCounter=0;BitCounter<8;BitCounter++) {
    nop();           // Delay50ns();
  	PCLK_HIGH;
    nop();           // Delay50ns();
    Data=Data<<1;
    if (PDO) Data++; // Data = Data | PDO ;
    nop();           // Delay50ns();
    PCLK_LOW;
  }
  nop();             // Delay50ns();
  PSEL_OFF; 
  return Data;
}

void CC1020_loadSettings() {
  UINT8 i;
  UINT32 freqa, freqb;
  for (i=0;i<0x27;i++) currentCC1020Config.registerValues[i] = DefaultConfig_19200[i];
  currentCC1020Config.registerValues[0x27] = ACK_TIMEOUT_19;
  // Adjust freqa, freqb:
  freqa = ( ((UINT32)currentCC1020Config.registerValues[0x04]) << 16) +
          ( ((UINT32)currentCC1020Config.registerValues[0x05]) << 8) +
          ( ((UINT32)currentCC1020Config.registerValues[0x06]) );
  freqb = ( ((UINT32)currentCC1020Config.registerValues[0x08]) << 16) +
          ( ((UINT32)currentCC1020Config.registerValues[0x09]) << 8) +
          ( ((UINT32)currentCC1020Config.registerValues[0x0A]) );
  currentCC1020Config.registerValues[0x04] = (freqa & 0x00FF0000) >> 16;
  currentCC1020Config.registerValues[0x05] = (freqa & 0x0000FF00) >> 8;
  currentCC1020Config.registerValues[0x06] = (freqa & 0x000000FF);
  currentCC1020Config.registerValues[0x08] = (freqb & 0x00FF0000) >> 16;
  currentCC1020Config.registerValues[0x09] = (freqb & 0x0000FF00) >> 8;
  currentCC1020Config.registerValues[0x0A] = (freqb & 0x000000FF);
}

void CC1020_writeSettings() {
  UINT8 i;  
  for (i=0;i<0x27;i++) currentCC1020Config.registerValues[i] = DefaultConfig_19200[i];
  currentCC1020Config.ackTimeout = ACK_TIMEOUT_19;
  for (i=0;i<0x20;i++) CC1020_writeRegister(i , currentCC1020Config.registerValues[i]);
  // same for every config, not clear if needed anyway..
  CC1020_writeRegister(0x20, 0x00);  // POWERDOWN    
  CC1020_writeRegister(0x21, 0x4d);  // 0x21,                   
  CC1020_writeRegister(0x22, 0x10);  // 0x22,                 
  CC1020_writeRegister(0x23, 0x06);  // 0x23,                    
  CC1020_writeRegister(0x24, 0x00);  // 0x24,                
  CC1020_writeRegister(0x25, 0x40);  // 0x25,                
  CC1020_writeRegister(0x26, 0x00);  // 0x26,                      
  CC1020_writeRegister(0x27, 0x00);  // 0x27,    
}

void CC1020_init() {
  CC1020_paPower = PA_POWER;
  CC1020_setupPD();
  CC1020_reset();
  CC1020_writeSettings() ; 
  CC1020_wakeupRX(RX_CURRENT);
  if (!CC1020_calibrate());
  CC1020_wakeupTX(TX_CURRENT);
  if (!CC1020_calibrate());
  CC1020_wakeupRX(RX_CURRENT);
  CC1020_setupRX(RX_CURRENT) ;
  LNA_POWER_ON;
}


void CC1020_reset() {
  // Reset CC1020
  CC1020_writeRegister(CC1020_MAIN, 0x0F&~0x01);
  // Bring CC1020 out of reset
  CC1020_writeRegister(CC1020_MAIN, 0x1F);
}
  
char CC1020_calibrate() {
  unsigned char tmp;
  unsigned int TimeOutCounter ;
  // Turn off PA to avoid spurs during calibration in TX mode
  CC1020_writeRegister(CC1020_PA_POWER,0x00);
  // Start calibration
  CC1020_writeRegister(CC1020_CALIBRATE,0xB5);
  System_delay(3); 
  while((tmp =(CC1020_readRegister(CC1020_STATUS)&CAL_COMPLETE)) == 0);
  System_delay(2); 
  // Monitor lock
  for(TimeOutCounter=LOCK_TIMEOUT; ((CC1020_readRegister(CC1020_STATUS)&0x10)==0)&&(TimeOutCounter>0); TimeOutCounter--);
  // Restore PA_POWER
  // WriteToCC1020Register(CC1020_PA_POWER,DefaultConfig[0x1C]);
  CC1020_writeRegister(CC1020_PA_POWER, CC1020_paPower);
  // Return state of LOCK_CONTINUOUS bit
  return ((CC1020_readRegister(CC1020_STATUS)&0x10)==0x10);
}

char CC1020_setupRX(char RXANALOG) {
  volatile int TimeOutCounter;
  char lock_status;
  // Switch into RX, switch to freq. reg A
  CC1020_writeRegister(CC1020_MAIN,0x11);
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,RXANALOG);
  System_delay(1); // Wait for 1 msec  
  // Monitor LOCK
  for(TimeOutCounter=LOCK_TIMEOUT; ((CC1020_readRegister(CC1020_STATUS)&LOCK_CONTINOUS)==0)&&(TimeOutCounter>0); TimeOutCounter--);
  // If PLL in lock
  if((CC1020_readRegister(CC1020_STATUS)&0x10)==LOCK_CONTINOUS) lock_status = LOCK_OK; // Indicate PLL in LOCK
  // Else (PLL out of LOCK)
  else {
    // If recalibration ok
    if(CC1020_calibrate()) lock_status = LOCK_RECAL_OK; // Indicate PLL in LOCK
	// Else (recalibration failed)
    else lock_status = LOCK_NOK; // Indicate PLL out of LOCK
  }
  // Switch RX part of CC1020 on
  CC1020_writeRegister(CC1020_MAIN,0x01);
  //WriteToCC1020Register(CC1020_INTERFACE,0x12);
  CC1020_writeRegister(CC1020_INTERFACE,0x02);
  // Return LOCK status to application
  return (lock_status);
}

char CC1020_setupTX(char TXANALOG) {
  volatile int TimeOutCounter;
  char lock_status;
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,TXANALOG);
  // Switch into TX, switch to freq. reg B
  CC1020_writeRegister(CC1020_MAIN,0xC1);
  System_delay(1); // Wait for 1 msec
  // Monitor LOCK, lasts 420 - 510 cycles @ 4505600 = 93 us - 113 us
  for(TimeOutCounter=LOCK_TIMEOUT; ((CC1020_readRegister(CC1020_STATUS)&0x10)==0)&&(TimeOutCounter>0); TimeOutCounter--);
  // If PLL in lock
  if((CC1020_readRegister(CC1020_STATUS)&0x10)==0x10) lock_status = LOCK_OK;  // Indicate PLL in LOCK
  // Else (PLL out of LOCK)
  else {
	// If recalibration ok
    if(CC1020_calibrate()) lock_status = LOCK_RECAL_OK; // Indicate PLL in LOCK
	// Else (recalibration failed)
    else lock_status = LOCK_NOK; // Indicate PLL out of LOCK
  }
  // Restore PA_POWER
  CC1020_writeRegister(CC1020_PA_POWER, CC1020_paPower);
  // Turn OFF DCLK squelch in TX
  CC1020_writeRegister(CC1020_INTERFACE, 0x01);
  // Return LOCK status to application
  return (lock_status);
}

void CC1020_setupPD() {
  // Put CC1020 into power-down
  CC1020_writeRegister(CC1020_MAIN,0x1F);
  // Turn off PA to minimise current draw
  CC1020_writeRegister(CC1020_PA_POWER,0x00);
}

void CC1020_wakeupRX(char RXANALOG) {
  // Turn on xtal oscillator core
  CC1020_writeRegister(CC1020_MAIN,0x1B);
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,RXANALOG);
  // Insert wait routine here, must wait for xtal oscillator to stabilise, 
  // typically takes 2-5ms.
  System_delay(3); // DelayMs(5); // NOT NEEDED?
  // Turn on bias generator
  CC1020_writeRegister(CC1020_MAIN,0x19);
  System_delay(1) ; // NOT NEEDED?
  // Turn on frequency synthesiser
  CC1020_writeRegister(CC1020_MAIN,0x11);
}

void CC1020_wakeupTX(char TXANALOG) {
  // Turn on xtal oscillator core
  CC1020_writeRegister(CC1020_MAIN,0xDB);
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,TXANALOG);
  // Insert wait routine here, must wait for xtal oscillator to stabilise, 
  // typically takes 2-5ms.
  System_delay(3); // DelayMs(5); // NOT NEEDED?
  // Turn on bias generator
  CC1020_writeRegister(CC1020_MAIN,0xD9);
  System_delay(1); // NOT NEEDED?
  // Turn on frequency synthesiser
  System_delay(1);
  CC1020_writeRegister(CC1020_MAIN,0xD1);
}

COMMAND(stp, 0) {
  CC1020_paPower = String_fromHexToByte((unsigned char*) &str[4]);
  Configuration.transmitPower = CC1020_paPower;
}

COMMAND(rtp, 0) {
  String_write(term_reply, "[rpa] ");
  String_fromByteToHex(CC1020_paPower, (UINT8*)(&term_reply[6]));
  term_reply[8] = '\0';
}

COMMAND(rss, 0) {
  UINT16 ta1, ta2, tb1, tb2;
  UINT16 myrssi;
  ta1 = TAR;
  tb1 = TBR;
  myrssi = CC1020_readRegister(CC1020_RSS);
  ta2 = TAR;
  tb2 = TBR;
  String_write(term_reply, "[rss] rssi=%i ta=%i tb=%i", myrssi, ta2-ta1, tb2-tb1);
}

COMMAND(rcs, 0) {
  UINT8 reg = CC1020_readRegister(CC1020_VGA4);
  UINT8 cs = reg & 0x1F;
  String_write(term_reply, "[rcs] CS in VGA4: %i (%b)", cs, reg);
}

COMMAND(scs, 0) {
  UINT8 set = String_parseInt(& str[4], NULL) & 0x1F;
  UINT8 reg = CC1020_readRegister(CC1020_VGA4);
  reg = (reg & 0xE0) | set;
  CC1020_writeRegister(CC1020_VGA4, reg);
}

COMMAND(rvg, 0) {
  UINT8 reg = CC1020_readRegister(CC1020_VGA3);
  UINT8 cs = reg & 0x1F;
  String_write(term_reply, "[rvg] VGS in VGA3: %i (%b)", cs, reg);
}

COMMAND(svg, 0) {
  UINT8 set = String_parseInt(& str[4], NULL) & 0x1F;
  UINT8 reg = CC1020_readRegister(CC1020_VGA3);
  reg = (reg & 0xE0) | set;
  CC1020_writeRegister(CC1020_VGA3, reg);
}
