/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include "ScatterWeb.System.h"

COMMAND(png, SERIALONLY) {
  packet_t packet;
  packet.from = Configuration.id;
  packet.to = String_parseInt(&str[4], NULL);   packet.type = PING_PACKET;
  packet.header_length = 0;                     packet.header = NULL;
  packet.data_length = 0;                       packet.data = NULL;
  Net_send(&packet, NULL);
}

// RST <br> Resets the MSP.
COMMAND(RST, 0) {
  System_reset();
}

// Prints current time in seconds,milliseconds and broken down time.
COMMAND(tim, 0) {
  btime_t btime;
  time_t time = CURRENT_TIME;
  Time_convertTime(&time, &btime);
  String_write(term_reply, "[tim] %l,%i %t", time.secs, time.millis, &btime);
}

// Sets current time as a long of seconds.
COMMAND(tis, 0) {
  time_t time;
  time.secs = String_parseLongInt(& str[4], NULL);
  time.millis = 0;
  Time_set(&time);
}

// Sets current time in broken down time format [yy-mm-dd hh:mm:ss].
COMMAND(tiz, 0) {
  btime_t btime;
  time_t time;
  Time_read(&str[4], &btime);
  Time_convertBtime(&btime, &time);
  Time_set(&time);
  Time_convertTime(&time, &btime);
}

// Clears all panics and does a reset.
COMMAND(pcl, 0) {
  System_reset();
}

// Lists all available commands.
COMMAND(lst, SERIALONLY) {
  extern command_t __term_start;  // __term_start is provided by the linker script, start of .commands segment.
  extern command_t __term_end;    // __term_end is provided by the linker script, end of .commands segment.
  command_t* c = &__term_start;
  while(c < &__term_end){ //and then loops through the command_t list
    Comm_print("%c%c%c(%x, %b)  ", c->code[0], c->code[1], c->code[2], c->function, c->flags);
    if((c-&__term_start)%5==4)
      Comm_print("\r\n");
    c++;
  }
  Comm_print("\r\n%i commands alltogether.\r\n", (&__term_end - &__term_start));
}

// Global term_reply var, generic search&execute function and serial and radio interface.
UINT8 term_reply[TERM_BUFSIZE]; 

// Search & execute a given command.
// If flag SERIAL_ONLY is set and a SERIAL_ONLY command is called nothing will
// be written to term_reply! If no flag is set term_reply will always contain
// some string after doCommand.
void _Messaging_doCommand(const UINT8* str, UINT8 flags) {
  extern command_t __term_start; // __term_start is provided by the linker script, start of .commands segment.
  extern command_t __term_end;   // __term_end is provided by the linker script, end of .commands segment.
  // The list of command_t structs automatically with the COMMAND Macro
  // begins at __term_start, which is provided by the linker script.
  command_t* c = &__term_start;
  // The algorithm first clears term_reply[0] 
  term_reply[0] = '\0';
  while(c < &__term_end) { // and then loops through the command_t list
    if( (str[0] == c->code[0]) && (str[1] == c->code[1]) && (str[2] == c->code[2]) ) { // until a matching command is found.
      if(c->flags & SERIALONLY) { // If it is serial only command
         // and SERIAL_ONLY in flags is set, the function is simply executed. This
         // is currently the only case when nothing is written to term_reply.
         if(flags & SERIALONLY) c->function(str);
         // Otherwise a "Not supported" answer is generated.
		 else String_write(term_reply, "[%c%c%c] NOT SUPPORTED (serial only)", str[0], str[1], str[2]);
      }
      else { // If it is a standard command (not serial only) which writes its reply to term_reply
        // the function is called.
        c->function(str);
        // Standard commands can leave term_reply untouched, if that is the case
        // a standard "OK" is generated.
        if(term_reply[0] == '\0') String_write(term_reply, "[%c%c%c] OK", str[0], str[1], str[2]);
      }
      return;
    }
    c++;       
  }
  // If no command exists a "UNKNOWN COMMAND" reply is generated.
  String_write(term_reply, "[%c%c%c] UNKNOWN COMMAND", str[0], str[1], str[2]);
}

void _Messaging_serialHandler(void* line) {
  UINT8* str = (UINT8*) line;
  if(str[0]== '\0') goto commclear;
  if(str[1]== '\0') goto commclear;
  if(str[2]== '\0') goto commclear;
  if(str[0]=='@') {
    packet_t packet;
	packet.from = Configuration.id;
    packet.to = String_parseInt(&str[1], &packet.data);
    packet.data++; // packet.data is now on first byte of remote terminal command
    for(packet.data_length=0; packet.data[packet.data_length] != '\0'; packet.data_length++);
    packet.data_length++;
    packet.type = MESSAGING_REQUEST_PACKET;
    packet.header = 0;
    packet.header_length = 0;
    Net_send(&packet, NULL);   
    goto commclear;
  } 
  else {
    _Messaging_doCommand(str, SERIALONLY);
    if(term_reply[0]!='\0') Comm_print("%s\r\n", term_reply);
    goto commclear;
  }
  commclear:  Comm_clear();
}

void _Messaging_requestHandler() {
  packet_t packet;
  packet.from = Configuration.id;
  packet.to = rxPacket.from;  
  packet.type = MESSAGING_REPLY_PACKET;
  packet.header = 0;   
  packet.header_length = 0;
  _Messaging_doCommand(rxPacket.data, 0);  
  // Count length and send data
  packet.data = term_reply;         
  for(packet.data_length=0; packet.data[packet.data_length] != '\0'; packet.data_length++);
  packet.data_length++;
  Net_send(&packet, 0);     
}

void _Messaging_replyHandler() {
  Comm_print("[%i] %s\r\n", rxPacket.from, rxPacket.data);
}

void Messaging_on() {
  System_registerCallback(C_SERIAL, _Messaging_serialHandler);        
}

bool Messaging_radioHandler() {
  if ((rxPacket.to == BROADCAST || rxPacket.to == Configuration.id)) {
    switch(rxPacket.type) {
      case MESSAGING_REQUEST_PACKET: _Messaging_requestHandler(); break;
      case MESSAGING_REPLY_PACKET: _Messaging_replyHandler(); break;
      default: break;
    }
  }
  return True;
}

// rfl xxxx xxxx <br> Reads flash contents like rer does for EEPROM.
// read flash (and other local memory)
COMMAND(rfl, SERIALONLY) { 
  UINT8 higha, lowa, eb;
  UINT8 *fromad, *toad, *pointer;
  higha = String_fromHexToByte( &str[4]);
  lowa = String_fromHexToByte( &str[6]);
  fromad = (UINT8*) ((higha << 8) + lowa);
  higha = String_fromHexToByte( &str[9]);
  lowa = String_fromHexToByte( &str[11]);
  toad = (UINT8*) ((higha << 8) + lowa);
  for(pointer=fromad; (pointer != toad+1); pointer++) {     
    if(((pointer - fromad) % 16) == 0) Comm_print("\r\n%x:", pointer);
    if( ((pointer - fromad) % 4) == 0) Comm_print("   ");
    eb = *pointer;
    Comm_print(" %b", eb);
  }
  Comm_print("\r\n");
}

COMMAND(rpc, 0) {
  String_write(term_reply, "[rpc] %x", READ_PC);
}
